import 'package:fl_chart/fl_chart.dart';
import 'package:healthrian_ble_interface/healthrian_ble_interface.dart';
import 'package:healthrian_ble_interface/src/extention/extension.dart';
import 'package:healthrian_common_support/healthrian_common_support.dart';
import 'package:lzma/lzma.dart';

typedef callbackPeriod = (
  void Function(Channels, DataFilter)? callback,
  int period
);

class ModuleEcgBuiltin {
  static List<List<double>>? udfConvert({
    required List<int> event,
    required int channelCount,
    required int sampleCount,
  }) {
    /// [channel_number] [sample_number]
    final res = List.generate(
      channelCount,
      (_) => List.filled(sampleCount, 0.0),
      growable: false,
    );
    if (event.isEmpty) return null;
    if (event.length < 4) return null;
    final id = event[3];
    if (id != 0x11) return null;

    /// [V6, L1, L2, V2, V3, V4, V5, V6, V1]
    final ordering = [11, 0, 1, 7, 8, 9, 10, 6];
    final skip = 6;
    for (var i = 0; i < sampleCount; i++) {
      for (var j = 0; j < 8; j++) {
        final msb = event[skip + 24 * i + j];
        final mid = event[skip + 24 * i + j + 1];
        final lsb = event[skip + 24 * i + j + 2];
        final value = toDouble(msb, mid, lsb);
        res[ordering[j]][i] = value;
      }
    }
    for (var i = 0; i < sampleCount; i++) {
      final l1 = res[0][i];
      final l2 = res[1][i];
      res[2].add(l3(l1, l2));
      res[3].add(aVR(l1, l2));
      res[4].add(aVL(l1, l2));
      res[5].add(aVF(l1, l2));
    }
    return res;
  }

  static Stream<List<int>>? udfLeadSource({
    required ModuleBle ble,
  }) {
    return ble.streams.eventEcg;
  }
}

class ModuleEcg {
  ModuleEcg({
    required this.ble,
    this.commands = const Commands(),
    this.udfConvert = ModuleEcgBuiltin.udfConvert,
    this.udfLeadSource = ModuleEcgBuiltin.udfLeadSource,
  });

  final ModuleBle ble;
  final Commands commands;

  final List<List<double>>? Function({
    required List<int> event,
    required int channelCount,
    required int sampleCount,
  }) udfConvert;

  final Stream<List<int>>? Function({
    required ModuleBle ble,
  }) udfLeadSource;

  final samplingRate = 250;
  final channelCount = 12;
  final sampleCount = 9;

  /// [maxBufferStreamEcg] decides a maximum length of buffered Stream.
  /// For 40 seconds buffer and 250 sampling rate,
  /// It uses 40 * 250 by default.
  int maxBufferStreamEcg = 40 * 250;

  /// [lengthBufferStreamEcg] decides a length of buffered Stream,
  /// after flushing buffer when reaching a maximum size.
  /// For 30 seconds buffer and 250 sampling rate,
  /// It uses 30 * 250 by default.
  int lengthBufferStreamEcg = 30 * 250;

  /// [screenBufferStreamingEcg] decides a length of buffered Stream,
  /// which is shown to screen.
  /// For 5 seconds buffer and 250 sampling rate,
  /// It uses 5 * 250 by default.
  int screenBufferStreamingEcg = 5 * 250;

  /// [refreshBufferEcg] decides a length to be hidden after refreshed.
  /// For 0.5 seconds range to be hidden and 250 sampling rate,
  /// It use 250 ~/ 2 by default.
  int refreshBufferEcg = 250 ~/ 2;

  /// [isFilteredEcg] decides whether ecg signal is filtered.
  bool isFilteredEcg = true;

  /// If [isCompressedEcg] is `true`,
  /// ecg signal is decompressed by LZMA algorithm.
  bool isCompressedEcg = false;

  DataFilter dataFilter = DataFilter.initialize(8);

  AdaptiveHR adaptiveHR = AdaptiveHR(0);

  List<callbackPeriod> callbacks = [
    (null, 7500),
    (null, 2500),
  ];

  /// [lastChannel] is a latest stream value from [ble.streams.bufferedEcg]
  Channels? lastChannel;

  bool _isPaused = false;
  bool _isMeasuring = false;

  ///////////////////////////////////////////////////////////////////////

  Future<void> startMeasurementEcg() async {
    dataFilter = DataFilter.initialize(8);
    adaptiveHR = AdaptiveHR(0);

    await _sendCommandSetTime();
    await _sendCommandReadEcg();

    _isMeasuring = true;
    _isPaused = false;

    ble.streams
      ..rawEcg = _getStream().broadcast
      ..decompressedEcg = _getDecompressedStream().broadcast
      ..convertedEcg = _getRecentStream().broadcast
      ..eventEcg = _events()?.broadcast
      ..leadStateEcg = _getLeadStateStream().broadcast
      ..heartRateEcg = _getHeartRateStream().broadcast
      ..bufferedEcg = _getBufferedStream().broadcast
      ..bufferedFilterEcg = _getBufferedFilterStream().broadcast
      ..plotEcg = _getBufferedPlotStream().broadcast
      ..plotRTEcg = _getBufferedPlotRTStream().broadcast;
  }

  /// [pauseMeasurementEcg]
  Future<void> pauseMeasurementEcg() async {
    _isPaused = !_isPaused;
  }

  /// [stopMeasurementEcg] sends stop_ecg_packet to device.
  Future<void> stopMeasurementEcg() async {
    _isMeasuring = false;
    final control = ble.points[Characteristics.STOP_ECG];
    await control.writePacket(commands.stopEcg);
  }

  Future<void> _sendCommandReadEcg() async {
    final control = ble.points[Characteristics.START_ECG];
    await control.writePacket(commands.startEcg);
  }

  Future<void> _sendCommandSetTime() async {
    final control = ble.points[Characteristics.SET_DATA];
    await control.writePacket(commands.setDate);
  }

  Stream<List<int>> _getStream() async* {
    final measurement = ble.points[Characteristics.MEASUREMENT_ECG];
    if (measurement == null) return;
    try {
      await measurement.setNotifyValue(true);
    } catch (e) {
      '_getStream : $e'.log;
    }

    await for (final res in measurement.lastValueStream) {
      if (!_isMeasuring || ble.device == null) break;
      if (_isPaused) continue;
      yield res;
    }
  }

  Stream<List<int>> _getDecompressedStream() async* {
    final events = ble.streams.rawEcg;
    if (events == null) return;
    await for (final event in events) {
      yield lzma.decode(event);
    }
  }

  Stream<List<int>>? _events() {
    return isCompressedEcg ? ble.streams.decompressedEcg : ble.streams.rawEcg;
  }

  ///
  Stream<List<List<double>>> _getRecentStream() async* {
    final events = ble.streams.eventEcg;
    if (events == null) return;
    await for (final event in events) {
      final value = udfConvert(
        event: event,
        channelCount: channelCount,
        sampleCount: sampleCount,
      );
      if (value == null) continue;
      yield value;
    }
  }

  ///
  Stream<LeadState?> _getLeadStateStream() async* {
    final events = udfLeadSource(
      ble: ble,
    );
    if (events == null) return;

    bool checkBit(int value, int bit) => (value & (1 << bit)) == 0;
    final leadState = LeadState.empty();

    var now = DateTime.now();

    await for (final event in events) {
      if (event.isEmpty) continue;
      if (event.length < 4) continue;
      final id = event[3];
      if (id != 0x11) continue;

      final currentTime = DateTime.now();
      if (currentTime.difference(now) <= const Duration(milliseconds: 500)) {
        continue;
      }

      now = currentTime;

      if (event.length < 6) continue;
      final leadStateHigh = event[4];
      final leadStateLow = event[5];

      yield leadState.copyWith(
        ra: checkBit(leadStateHigh, 1),
        ll: checkBit(leadStateLow, 2),
        la: checkBit(leadStateLow, 1),
        v1: checkBit(leadStateLow, 7),
        v2: checkBit(leadStateLow, 3),
        v3: checkBit(leadStateLow, 4),
        v4: checkBit(leadStateLow, 5),
        v5: checkBit(leadStateLow, 6),
        v6: checkBit(leadStateLow, 0),
      );
    }
    yield null;
  }

  ///
  Stream<int?> _getHeartRateStream() async* {
    final events = ble.streams.eventEcg;
    if (events == null) return;

    var lastSentTime = DateTime.now();

    await for (final event in events) {
      if (event.isEmpty) continue;
      if (event.length < 6) continue;
      final id = event[3];
      if (id != 0x11) continue;

      final itemCount = event.length ~/ 3;

      for (var i = 2; i < itemCount; i++) {
        if (i % 8 != 0) continue;

        final raw = toDouble(
          event[i * 3 + 0],
          event[i * 3 + 1],
          event[i * 3 + 2],
        );

        final output = adaptiveHR.calculateHR(
          raw,
          true,
        );

        final hr = output <= 0 ? null : output;
        if (hr == null &&
            DateTime.now().difference(lastSentTime) < Duration(seconds: 2))
          continue;

        yield hr;
        lastSentTime = DateTime.now();
      }
    }
    yield null;
  }

  Stream<Channels> _getBufferedStream() async* {
    final events = ble.streams.convertedEcg;
    if (events == null) return;

    final res = List.generate(channelCount, (index) => <double>[]);

    var x = 0.0;
    final callbacksTargets = [for (final callback in callbacks) callback.$2];

    await for (final event in events) {
      for (var i = 0; i < channelCount; i++) {
        res[i].addAll(event[i]);
      }

      for (final channel in res) {
        if (channel.length > maxBufferStreamEcg) {
          channel.removeRange(0, channel.length - lengthBufferStreamEcg);
        }
      }

      final channels = Channels.fromList(res);
      yield channels;

      for (var i = 0; i < callbacks.length; i++) {
        final callback = callbacks[i].$1;
        final period = callbacks[i].$2;
        if (callback != null && x >= callbacksTargets[i]) {
          callback.call(channels, dataFilter);
          callbacksTargets[i] += period;
        }
      }

      lastChannel = channels;

      x += sampleCount;
    }
  }

  Stream<Channels> _getBufferedFilterStream() async* {
    final events = ble.streams.bufferedEcg;
    if (events == null) return;

    await for (final event in events) {
      for (var i = event.l2.length - sampleCount; i < event.l2.length; i++) {
        final l1 = dataFilter.applyFilter(0, event.l1[i], true);
        final l2 = dataFilter.applyFilter(1, event.l2[i], true);

        event.l1[i] = l1;
        event.l2[i] = l2;
        event.l3[i] = l3(l1, l2);

        event.aVR[i] = aVR(l1, l2);
        event.aVL[i] = aVL(l1, l2);
        event.aVF[i] = aVF(l1, l2);

        event.v1[i] = dataFilter.applyFilter(2, event.v1[i], true);
        event.v2[i] = dataFilter.applyFilter(3, event.v2[i], true);
        event.v3[i] = dataFilter.applyFilter(4, event.v3[i], true);
        event.v4[i] = dataFilter.applyFilter(5, event.v4[i], true);
        event.v5[i] = dataFilter.applyFilter(6, event.v5[i], true);
        event.v6[i] = dataFilter.applyFilter(7, event.v6[i], true);
      }
      yield event;
    }
  }

  Stream<PlotChannels> _getBufferedPlotStream() async* {
    final events =
        isFilteredEcg ? ble.streams.bufferedFilterEcg : ble.streams.bufferedEcg;

    if (events == null) return;

    final resList = List.generate(channelCount, (index) => <FlSpot>[]);
    final res = PlotChannels.fromList(resList, 0);
    var x = 0.0;

    await for (final event in events) {
      for (var i = event.l2.length - sampleCount; i < event.l2.length; i++) {
        final j = i - event.l2.length + sampleCount;

        res.l1.add(FlSpot(x + j, event.l1[i]));
        res.l2.add(FlSpot(x + j, event.l2[i]));
        res.l3.add(FlSpot(x + j, event.l3[i]));

        res.aVR.add(FlSpot(x + j, event.aVR[i]));
        res.aVL.add(FlSpot(x + j, event.aVL[i]));
        res.aVF.add(FlSpot(x + j, event.aVF[i]));

        res.v1.add(FlSpot(x + j, event.v1[i]));
        res.v2.add(FlSpot(x + j, event.v2[i]));
        res.v3.add(FlSpot(x + j, event.v3[i]));
        res.v4.add(FlSpot(x + j, event.v4[i]));
        res.v5.add(FlSpot(x + j, event.v5[i]));
        res.v6.add(FlSpot(x + j, event.v6[i]));
      }

      if (res.l2.length > event.l2.length) {
        res.l1.removeRange(0, res.l1.length - event.l1.length);
        res.l2.removeRange(0, res.l2.length - event.l2.length);
        res.l3.removeRange(0, res.l3.length - event.l3.length);

        res.aVR.removeRange(0, res.aVR.length - event.aVR.length);
        res.aVL.removeRange(0, res.aVL.length - event.aVL.length);
        res.aVF.removeRange(0, res.aVF.length - event.aVF.length);

        res.v1.removeRange(0, res.v1.length - event.v1.length);
        res.v2.removeRange(0, res.v2.length - event.v2.length);
        res.v3.removeRange(0, res.v3.length - event.v3.length);
        res.v4.removeRange(0, res.v4.length - event.v4.length);
        res.v5.removeRange(0, res.v5.length - event.v5.length);
        res.v6.removeRange(0, res.v6.length - event.v6.length);
      }

      yield res;
      x += sampleCount;
    }
  }

  Stream<PlotChannels?> _getBufferedPlotRTStream() async* {
    final events = ble.streams.plotEcg;

    if (events == null) return;

    double skip(double x) {
      final newX = x % screenBufferStreamingEcg;
      final newXSkip = newX - newX;
      return newXSkip;
    }

    FlSpot f(FlSpot original) {
      return FlSpot(
        skip(original.x),
        original.y,
      );
    }

    final resList = List.generate(
      channelCount,
      (_) => List<FlSpot>.generate(
        screenBufferStreamingEcg,
        (index) => FlSpot.nullSpot,
        growable: false,
      ),
    );

    final res = PlotChannels.fromList(resList, 0);
    var x = 0;

    await for (final event in events) {
      for (var i = event.l2.length - sampleCount; i < event.l2.length; i++) {
        final j = i - event.l2.length + sampleCount;

        final index = skip(x + j.toDouble()).round();

        res.l1[index] = f(event.l1[i]);
        res.l2[index] = f(event.l2[i]);
        res.l3[index] = f(event.l3[i]);

        res.aVR[index] = f(event.aVR[i]);
        res.aVL[index] = f(event.aVL[i]);
        res.aVF[index] = f(event.aVF[i]);

        res.v1[index] = f(event.v1[i]);
        res.v2[index] = f(event.v2[i]);
        res.v3[index] = f(event.v3[i]);
        res.v4[index] = f(event.v4[i]);
        res.v5[index] = f(event.v5[i]);
        res.v6[index] = f(event.v6[i]);

        if (refreshBufferEcg > 0) {
          final refreshedIndex =
              skip(x + j.toDouble() + refreshBufferEcg).round();

          res.l1[refreshedIndex] = FlSpot.nullSpot;
          res.l2[refreshedIndex] = FlSpot.nullSpot;
          res.l3[refreshedIndex] = FlSpot.nullSpot;

          res.aVR[refreshedIndex] = FlSpot.nullSpot;
          res.aVL[refreshedIndex] = FlSpot.nullSpot;
          res.aVF[refreshedIndex] = FlSpot.nullSpot;

          res.v1[refreshedIndex] = FlSpot.nullSpot;
          res.v2[refreshedIndex] = FlSpot.nullSpot;
          res.v3[refreshedIndex] = FlSpot.nullSpot;
          res.v4[refreshedIndex] = FlSpot.nullSpot;
          res.v5[refreshedIndex] = FlSpot.nullSpot;
          res.v6[refreshedIndex] = FlSpot.nullSpot;
        }
      }

      x += sampleCount;

      yield res.copyWith(
        pivot: (x % screenBufferStreamingEcg).toDouble(),
      );
    }
    yield null;
  }
}
